  (define b1 (make-button '(-text "button1")))
  (define b2 (make-button '(-text "button2")))
  (define b3 (make-button '(-text "button3")))
  (define b4 (make-button '(-text "button4")))
  (define b5 (make-button '(-text "button5")))
  (define b6 (make-button '(-text "very very long button6")))
  (define b7 (make-button '(-text "b7")))
  (define b8 (make-button '(-text "b8")))
  (define b9 (make-button '(-text "b9")))
  (define s (make-space))

  (define a (make-application "test"))
  (define ab (make-array-box (list b1 'left b3)
			     (list 'up 'up b6)
			     (list b7 b8 b9)))
  (add-child! a ab)

(define (test2)
  (define c (make-canvas '(-width 300 -height 300 -background white)))
  (define b1 (make-button '(-text "button1" -background red)))
  (define b2 (make-button '(-text "button2" -background red)))
  (define b3 (make-button '(-text "button3" -background red)))
  (define s1 (make-space '(-background red)))
  (define s2 (make-space '(-background red)))
  (define hb (make-hbox s1 b1 b2 s2 b3))
  (define vb (make-vbox c hb))
  (define a (make-application "test"))
  (add-child! a vb)
  a)

#|
(define c (make-canvas '(-width 200 -height 200)))
(define b1 (make-button '(-text "button1")))
(define b2 (make-button '(-text "button2")))
(define box1 (make-hbox b1 b2))
(define box2 (make-vbox c box1))
(define a (make-application "test"))
(add-child! a box2)

|#