(define-constant X_PROTOCOL 11)
(define-constant X_PROTOCOL_REVISION 0)
(define-constant NONE 0)
(define-constant PARENTRELATIVE 1)
(define-constant COPYFROMPARENT 0)
(define-constant POINTERWINDOW 0)
(define-constant INPUTFOCUS 1)
(define-constant POINTERROOT 1)
(define-constant ANYPROPERTYTYPE 0)
(define-constant ANYKEY 0)
(define-constant ANYBUTTON 0)
(define-constant ALLTEMPORARY 0)
(define-constant CURRENTTIME 0)
(define-constant NOSYMBOL 0)
(define-constant NOEVENTMASK 0)
(define-constant KEYPRESSMASK 1)
(define-constant KEYRELEASEMASK 2)
(define-constant BUTTONPRESSMASK 4)
(define-constant BUTTONRELEASEMASK 8)
(define-constant ENTERWINDOWMASK 16)
(define-constant LEAVEWINDOWMASK 32)
(define-constant POINTERMOTIONMASK 64)
(define-constant POINTERMOTIONHINTMASK 128)
(define-constant BUTTON1MOTIONMASK 256)
(define-constant BUTTON2MOTIONMASK 512)
(define-constant BUTTON3MOTIONMASK 1024)
(define-constant BUTTON4MOTIONMASK 2048)
(define-constant BUTTON5MOTIONMASK 4096)
(define-constant BUTTONMOTIONMASK 8192)
(define-constant KEYMAPSTATEMASK 16384)
(define-constant EXPOSUREMASK 32768)
(define-constant VISIBILITYCHANGEMASK 65536)
(define-constant STRUCTURENOTIFYMASK 131072)
(define-constant RESIZEREDIRECTMASK 262144)
(define-constant SUBSTRUCTURENOTIFYMASK 524288)
(define-constant SUBSTRUCTUREREDIRECTMASK 1048576)
(define-constant FOCUSCHANGEMASK 2097152)
(define-constant PROPERTYCHANGEMASK 4194304)
(define-constant COLORMAPCHANGEMASK 8388608)
(define-constant OWNERGRABBUTTONMASK 16777216)
(define-constant KEYPRESS 2)
(define-constant KEYRELEASE 3)
(define-constant BUTTONPRESS 4)
(define-constant BUTTONRELEASE 5)
(define-constant MOTIONNOTIFY 6)
(define-constant ENTERNOTIFY 7)
(define-constant LEAVENOTIFY 8)
(define-constant FOCUSIN 9)
(define-constant FOCUSOUT 10)
(define-constant KEYMAPNOTIFY 11)
(define-constant EXPOSE 12)
(define-constant GRAPHICSEXPOSE 13)
(define-constant NOEXPOSE 14)
(define-constant VISIBILITYNOTIFY 15)
(define-constant CREATENOTIFY 16)
(define-constant DESTROYNOTIFY 17)
(define-constant UNMAPNOTIFY 18)
(define-constant MAPNOTIFY 19)
(define-constant MAPREQUEST 20)
(define-constant REPARENTNOTIFY 21)
(define-constant CONFIGURENOTIFY 22)
(define-constant CONFIGUREREQUEST 23)
(define-constant GRAVITYNOTIFY 24)
(define-constant RESIZEREQUEST 25)
(define-constant CIRCULATENOTIFY 26)
(define-constant CIRCULATEREQUEST 27)
(define-constant PROPERTYNOTIFY 28)
(define-constant SELECTIONCLEAR 29)
(define-constant SELECTIONREQUEST 30)
(define-constant SELECTIONNOTIFY 31)
(define-constant COLORMAPNOTIFY 32)
(define-constant CLIENTMESSAGE 33)
(define-constant MAPPINGNOTIFY 34)
(define-constant LASTEVENT 35)
(define-constant SHIFTMASK 1)
(define-constant LOCKMASK 2)
(define-constant CONTROLMASK 4)
(define-constant MOD1MASK 8)
(define-constant MOD2MASK 16)
(define-constant MOD3MASK 32)
(define-constant MOD4MASK 64)
(define-constant MOD5MASK 128)
(define-constant SHIFTMAPINDEX 0)
(define-constant LOCKMAPINDEX 1)
(define-constant CONTROLMAPINDEX 2)
(define-constant MOD1MAPINDEX 3)
(define-constant MOD2MAPINDEX 4)
(define-constant MOD3MAPINDEX 5)
(define-constant MOD4MAPINDEX 6)
(define-constant MOD5MAPINDEX 7)
(define-constant BUTTON1MASK 256)
(define-constant BUTTON2MASK 512)
(define-constant BUTTON3MASK 1024)
(define-constant BUTTON4MASK 2048)
(define-constant BUTTON5MASK 4096)
(define-constant ANYMODIFIER 32768)
(define-constant BUTTON1 1)
(define-constant BUTTON2 2)
(define-constant BUTTON3 3)
(define-constant BUTTON4 4)
(define-constant BUTTON5 5)
(define-constant NOTIFYNORMAL 0)
(define-constant NOTIFYGRAB 1)
(define-constant NOTIFYUNGRAB 2)
(define-constant NOTIFYWHILEGRABBED 3)
(define-constant NOTIFYHINT 1)
(define-constant NOTIFYANCESTOR 0)
(define-constant NOTIFYVIRTUAL 1)
(define-constant NOTIFYINFERIOR 2)
(define-constant NOTIFYNONLINEAR 3)
(define-constant NOTIFYNONLINEARVIRTUAL 4)
(define-constant NOTIFYPOINTER 5)
(define-constant NOTIFYPOINTERROOT 6)
(define-constant NOTIFYDETAILNONE 7)
(define-constant VISIBILITYUNOBSCURED 0)
(define-constant VISIBILITYPARTIALLYOBSCURED 1)
(define-constant VISIBILITYFULLYOBSCURED 2)
(define-constant PLACEONTOP 0)
(define-constant PLACEONBOTTOM 1)
(define-constant FAMILYINTERNET 0)
(define-constant FAMILYDECNET 1)
(define-constant FAMILYCHAOS 2)
(define-constant PROPERTYNEWVALUE 0)
(define-constant PROPERTYDELETE 1)
(define-constant COLORMAPUNINSTALLED 0)
(define-constant COLORMAPINSTALLED 1)
(define-constant GRABMODESYNC 0)
(define-constant GRABMODEASYNC 1)
(define-constant GRABSUCCESS 0)
(define-constant ALREADYGRABBED 1)
(define-constant GRABINVALIDTIME 2)
(define-constant GRABNOTVIEWABLE 3)
(define-constant GRABFROZEN 4)
(define-constant ASYNCPOINTER 0)
(define-constant SYNCPOINTER 1)
(define-constant REPLAYPOINTER 2)
(define-constant ASYNCKEYBOARD 3)
(define-constant SYNCKEYBOARD 4)
(define-constant REPLAYKEYBOARD 5)
(define-constant ASYNCBOTH 6)
(define-constant SYNCBOTH 7)
(define-constant REVERTTONONE 0)
(define-constant REVERTTOPOINTERROOT 1)
(define-constant REVERTTOPARENT 2)
(define-constant SUCCESS 0)
(define-constant BADREQUEST 1)
(define-constant BADVALUE 2)
(define-constant BADWINDOW 3)
(define-constant BADPIXMAP 4)
(define-constant BADATOM 5)
(define-constant BADCURSOR 6)
(define-constant BADFONT 7)
(define-constant BADMATCH 8)
(define-constant BADDRAWABLE 9)
(define-constant BADACCESS 10)
(define-constant BADALLOC 11)
(define-constant BADCOLOR 12)
(define-constant BADGC 13)
(define-constant BADIDCHOICE 14)
(define-constant BADNAME 15)
(define-constant BADLENGTH 16)
(define-constant BADIMPLEMENTATION 17)
(define-constant FIRSTEXTENSIONERROR 128)
(define-constant LASTEXTENSIONERROR 255)
(define-constant INPUTOUTPUT 1)
(define-constant INPUTONLY 2)
(define-constant CWBACKPIXMAP 1)
(define-constant CWBACKPIXEL 2)
(define-constant CWBORDERPIXMAP 4)
(define-constant CWBORDERPIXEL 8)
(define-constant CWBITGRAVITY 16)
(define-constant CWWINGRAVITY 32)
(define-constant CWBACKINGSTORE 64)
(define-constant CWBACKINGPLANES 128)
(define-constant CWBACKINGPIXEL 256)
(define-constant CWOVERRIDEREDIRECT 512)
(define-constant CWSAVEUNDER 1024)
(define-constant CWEVENTMASK 2048)
(define-constant CWDONTPROPAGATE 4096)
(define-constant CWCOLORMAP 8192)
(define-constant CWCURSOR 16384)
(define-constant CWX 1)
(define-constant CWY 2)
(define-constant CWWIDTH 4)
(define-constant CWHEIGHT 8)
(define-constant CWBORDERWIDTH 16)
(define-constant CWSIBLING 32)
(define-constant CWSTACKMODE 64)
(define-constant FORGETGRAVITY 0)
(define-constant NORTHWESTGRAVITY 1)
(define-constant NORTHGRAVITY 2)
(define-constant NORTHEASTGRAVITY 3)
(define-constant WESTGRAVITY 4)
(define-constant CENTERGRAVITY 5)
(define-constant EASTGRAVITY 6)
(define-constant SOUTHWESTGRAVITY 7)
(define-constant SOUTHGRAVITY 8)
(define-constant SOUTHEASTGRAVITY 9)
(define-constant STATICGRAVITY 10)
(define-constant UNMAPGRAVITY 0)
(define-constant NOTUSEFUL 0)
(define-constant WHENMAPPED 1)
(define-constant ALWAYS 2)
(define-constant ISUNMAPPED 0)
(define-constant ISUNVIEWABLE 1)
(define-constant ISVIEWABLE 2)
(define-constant SETMODEINSERT 0)
(define-constant SETMODEDELETE 1)
(define-constant DESTROYALL 0)
(define-constant RETAINPERMANENT 1)
(define-constant RETAINTEMPORARY 2)
(define-constant ABOVE 0)
(define-constant BELOW 1)
(define-constant TOPIF 2)
(define-constant BOTTOMIF 3)
(define-constant OPPOSITE 4)
(define-constant RAISELOWEST 0)
(define-constant LOWERHIGHEST 1)
(define-constant PROPMODEREPLACE 0)
(define-constant PROPMODEPREPEND 1)
(define-constant PROPMODEAPPEND 2)
(define-constant GXCLEAR 0)
(define-constant GXAND 1)
(define-constant GXANDREVERSE 2)
(define-constant GXCOPY 3)
(define-constant GXANDINVERTED 4)
(define-constant GXNOOP 5)
(define-constant GXXOR 6)
(define-constant GXOR 7)
(define-constant GXNOR 8)
(define-constant GXEQUIV 9)
(define-constant GXINVERT 10)
(define-constant GXORREVERSE 11)
(define-constant GXCOPYINVERTED 12)
(define-constant GXORINVERTED 13)
(define-constant GXNAND 14)
(define-constant GXSET 15)
(define-constant LINESOLID 0)
(define-constant LINEONOFFDASH 1)
(define-constant LINEDOUBLEDASH 2)
(define-constant CAPNOTLAST 0)
(define-constant CAPBUTT 1)
(define-constant CAPROUND 2)
(define-constant CAPPROJECTING 3)
(define-constant JOINMITER 0)
(define-constant JOINROUND 1)
(define-constant JOINBEVEL 2)
(define-constant FILLSOLID 0)
(define-constant FILLTILED 1)
(define-constant FILLSTIPPLED 2)
(define-constant FILLOPAQUESTIPPLED 3)
(define-constant EVENODDRULE 0)
(define-constant WINDINGRULE 1)
(define-constant CLIPBYCHILDREN 0)
(define-constant INCLUDEINFERIORS 1)
(define-constant UNSORTED 0)
(define-constant YSORTED 1)
(define-constant YXSORTED 2)
(define-constant YXBANDED 3)
(define-constant COORDMODEORIGIN 0)
(define-constant COORDMODEPREVIOUS 1)
(define-constant COMPLEX 0)
(define-constant NONCONVEX 1)
(define-constant CONVEX 2)
(define-constant ARCCHORD 0)
(define-constant ARCPIESLICE 1)
(define-constant GCFUNCTION 1)
(define-constant GCPLANEMASK 2)
(define-constant GCFOREGROUND 4)
(define-constant GCBACKGROUND 8)
(define-constant GCLINEWIDTH 16)
(define-constant GCLINESTYLE 32)
(define-constant GCCAPSTYLE 64)
(define-constant GCJOINSTYLE 128)
(define-constant GCFILLSTYLE 256)
(define-constant GCFILLRULE 512)
(define-constant GCTILE 1024)
(define-constant GCSTIPPLE 2048)
(define-constant GCTILESTIPXORIGIN 4096)
(define-constant GCTILESTIPYORIGIN 8192)
(define-constant GCFONT 16384)
(define-constant GCSUBWINDOWMODE 32768)
(define-constant GCGRAPHICSEXPOSURES 65536)
(define-constant GCCLIPXORIGIN 131072)
(define-constant GCCLIPYORIGIN 262144)
(define-constant GCCLIPMASK 524288)
(define-constant GCDASHOFFSET 1048576)
(define-constant GCDASHLIST 2097152)
(define-constant GCARCMODE 4194304)
(define-constant GCLASTBIT 22)
(define-constant FONTLEFTTORIGHT 0)
(define-constant FONTRIGHTTOLEFT 1)
(define-constant FONTCHANGE 255)
(define-constant XYBITMAP 0)
(define-constant XYPIXMAP 1)
(define-constant ZPIXMAP 2)
(define-constant ALLOCNONE 0)
(define-constant ALLOCALL 1)
(define-constant DORED 1)
(define-constant DOGREEN 2)
(define-constant DOBLUE 4)
(define-constant CURSORSHAPE 0)
(define-constant TILESHAPE 1)
(define-constant STIPPLESHAPE 2)
(define-constant AUTOREPEATMODEOFF 0)
(define-constant AUTOREPEATMODEON 1)
(define-constant AUTOREPEATMODEDEFAULT 2)
(define-constant LEDMODEOFF 0)
(define-constant LEDMODEON 1)
(define-constant KBKEYCLICKPERCENT 1)
(define-constant KBBELLPERCENT 2)
(define-constant KBBELLPITCH 4)
(define-constant KBBELLDURATION 8)
(define-constant KBLED 16)
(define-constant KBLEDMODE 32)
(define-constant KBKEY 64)
(define-constant KBAUTOREPEATMODE 128)
(define-constant MAPPINGSUCCESS 0)
(define-constant MAPPINGBUSY 1)
(define-constant MAPPINGFAILED 2)
(define-constant MAPPINGMODIFIER 0)
(define-constant MAPPINGKEYBOARD 1)
(define-constant MAPPINGPOINTER 2)
(define-constant DONTPREFERBLANKING 0)
(define-constant PREFERBLANKING 1)
(define-constant DEFAULTBLANKING 2)
(define-constant DISABLESCREENSAVER 0)
(define-constant DISABLESCREENINTERVAL 0)
(define-constant DONTALLOWEXPOSURES 0)
(define-constant ALLOWEXPOSURES 1)
(define-constant DEFAULTEXPOSURES 2)
(define-constant SCREENSAVERRESET 0)
(define-constant SCREENSAVERACTIVE 1)
(define-constant HOSTINSERT 0)
(define-constant HOSTDELETE 1)
(define-constant ENABLEACCESS 1)
(define-constant DISABLEACCESS 0)
(define-constant STATICGRAY 0)
(define-constant GRAYSCALE 1)
(define-constant STATICCOLOR 2)
(define-constant PSEUDOCOLOR 3)
(define-constant TRUECOLOR 4)
(define-constant DIRECTCOLOR 5)
(define-constant LSBFIRST 0)
(define-constant MSBFIRST 1)
(define-constant XA_PRIMARY 1)
(define-constant XA_SECONDARY 2)
(define-constant XA_ARC 3)
(define-constant XA_ATOM 4)
(define-constant XA_BITMAP 5)
(define-constant XA_CARDINAL 6)
(define-constant XA_COLORMAP 7)
(define-constant XA_CURSOR 8)
(define-constant XA_CUT_BUFFER0 9)
(define-constant XA_CUT_BUFFER1 10)
(define-constant XA_CUT_BUFFER2 11)
(define-constant XA_CUT_BUFFER3 12)
(define-constant XA_CUT_BUFFER4 13)
(define-constant XA_CUT_BUFFER5 14)
(define-constant XA_CUT_BUFFER6 15)
(define-constant XA_CUT_BUFFER7 16)
(define-constant XA_DRAWABLE 17)
(define-constant XA_FONT 18)
(define-constant XA_INTEGER 19)
(define-constant XA_PIXMAP 20)
(define-constant XA_POINT 21)
(define-constant XA_RECTANGLE 22)
(define-constant XA_RESOURCE_MANAGER 23)
(define-constant XA_RGB_COLOR_MAP 24)
(define-constant XA_RGB_BEST_MAP 25)
(define-constant XA_RGB_BLUE_MAP 26)
(define-constant XA_RGB_DEFAULT_MAP 27)
(define-constant XA_RGB_GRAY_MAP 28)
(define-constant XA_RGB_GREEN_MAP 29)
(define-constant XA_RGB_RED_MAP 30)
(define-constant XA_STRING 31)
(define-constant XA_VISUALID 32)
(define-constant XA_WINDOW 33)
(define-constant XA_WM_COMMAND 34)
(define-constant XA_WM_HINTS 35)
(define-constant XA_WM_CLIENT_MACHINE 36)
(define-constant XA_WM_ICON_NAME 37)
(define-constant XA_WM_ICON_SIZE 38)
(define-constant XA_WM_NAME 39)
(define-constant XA_WM_NORMAL_HINTS 40)
(define-constant XA_WM_SIZE_HINTS 41)
(define-constant XA_WM_ZOOM_HINTS 42)
(define-constant XA_MIN_SPACE 43)
(define-constant XA_NORM_SPACE 44)
(define-constant XA_MAX_SPACE 45)
(define-constant XA_END_SPACE 46)
(define-constant XA_SUPERSCRIPT_X 47)
(define-constant XA_SUPERSCRIPT_Y 48)
(define-constant XA_SUBSCRIPT_X 49)
(define-constant XA_SUBSCRIPT_Y 50)
(define-constant XA_UNDERLINE_POSITION 51)
(define-constant XA_UNDERLINE_THICKNESS 52)
(define-constant XA_STRIKEOUT_ASCENT 53)
(define-constant XA_STRIKEOUT_DESCENT 54)
(define-constant XA_ITALIC_ANGLE 55)
(define-constant XA_X_HEIGHT 56)
(define-constant XA_QUAD_WIDTH 57)
(define-constant XA_WEIGHT 58)
(define-constant XA_POINT_SIZE 59)
(define-constant XA_RESOLUTION 60)
(define-constant XA_COPYRIGHT 61)
(define-constant XA_NOTICE 62)
(define-constant XA_FONT_NAME 63)
(define-constant XA_FAMILY_NAME 64)
(define-constant XA_FULL_NAME 65)
(define-constant XA_CAP_HEIGHT 66)
(define-constant XA_WM_CLASS 67)
(define-constant XA_WM_TRANSIENT_FOR 68)
(define-constant XA_LAST_PREDEFINED 68)
(define-constant QUEUEDALREADY 0)
(define-constant QUEUEDAFTERREADING 1)
(define-constant QUEUEDAFTERFLUSH 2)
(define-constant ALLPLANES -1)
(define-constant NULL 0)
(define-constant NULLQUARK 0)
(define-constant NULLSTRING 0)
(define-constant XRMBINDTIGHTLY 0)
(define-constant XRMBINDLOOSELY 1)
(define-constant XRMOPTIONNOARG 0)
(define-constant XRMOPTIONISARG 1)
(define-constant XRMOPTIONSTICKYARG 2)
(define-constant XRMOPTIONSEPARG 3)
(define-constant XRMOPTIONRESARG 4)
(define-constant XRMOPTIONSKIPARG 5)
(define-constant XRMOPTIONSKIPLINE 6)
(define-constant NOVALUE 0)
(define-constant XVALUE 1)
(define-constant YVALUE 2)
(define-constant WIDTHVALUE 4)
(define-constant HEIGHTVALUE 8)
(define-constant ALLVALUES 15)
(define-constant XNEGATIVE 16)
(define-constant YNEGATIVE 32)
(define-constant USPOSITION 1)
(define-constant USSIZE 2)
(define-constant PPOSITION 4)
(define-constant PSIZE 8)
(define-constant PMINSIZE 16)
(define-constant PMAXSIZE 32)
(define-constant PRESIZEINC 64)
(define-constant PASPECT 128)
(define-constant PALLHINTS 252)
(define-constant INPUTHINT 1)
(define-constant STATEHINT 2)
(define-constant ICONPIXMAPHINT 4)
(define-constant ICONWINDOWHINT 8)
(define-constant ICONPOSITIONHINT 16)
(define-constant ICONMASKHINT 32)
(define-constant WINDOWGROUPHINT 64)
(define-constant ALLHINTS 127)
(define-constant DONTCARESTATE 0)
(define-constant NORMALSTATE 1)
(define-constant ZOOMSTATE 2)
(define-constant ICONICSTATE 3)
(define-constant INACTIVESTATE 4)
(define-constant RECTANGLEOUT 0)
(define-constant RECTANGLEIN 1)
(define-constant RECTANGLEPART 2)
(define-constant VISUALNOMASK 0)
(define-constant VISUALIDMASK 1)
(define-constant VISUALSCREENMASK 2)
(define-constant VISUALDEPTHMASK 4)
(define-constant VISUALCLASSMASK 8)
(define-constant VISUALREDMASKMASK 16)
(define-constant VISUALGREENMASKMASK 32)
(define-constant VISUALBLUEMASKMASK 64)
(define-constant VISUALCOLORMAPSIZEMASK 128)
(define-constant VISUALBITSPERRGBMASK 256)
(define-constant VISUALALLMASK 511)
(define-constant BITMAPSUCCESS 0)
(define-constant BITMAPOPENFAILED 1)
(define-constant BITMAPFILEINVALID 2)
(define-constant BITMAPNOMEMORY 3)
(define-constant XCSUCCESS 0)
(define-constant XCNOMEM 1)
(define-constant XCNOENT 2)
(define-constant XK_BACKSPACE 65288)
(define-constant XK_TAB 65289)
(define-constant XK_LINEFEED 65290)
(define-constant XK_CLEAR 65291)
(define-constant XK_RETURN 65293)
(define-constant XK_PAUSE 65299)
(define-constant XK_ESCAPE 65307)
(define-constant XK_DELETE 65535)
(define-constant XK_MULTI_KEY 65312)
(define-constant XK_KANJI 65313)
(define-constant XK_HOME 65360)
(define-constant XK_LEFT 65361)
(define-constant XK_UP 65362)
(define-constant XK_RIGHT 65363)
(define-constant XK_DOWN 65364)
(define-constant XK_PRIOR 65365)
(define-constant XK_NEXT 65366)
(define-constant XK_END 65367)
(define-constant XK_BEGIN 65368)
(define-constant XK_SELECT 65376)
(define-constant XK_PRINT 65377)
(define-constant XK_EXECUTE 65378)
(define-constant XK_INSERT 65379)
(define-constant XK_UNDO 65381)
(define-constant XK_REDO 65382)
(define-constant XK_MENU 65383)
(define-constant XK_FIND 65384)
(define-constant XK_CANCEL 65385)
(define-constant XK_HELP 65386)
(define-constant XK_BREAK 65387)
(define-constant XK_MODE_SWITCH 65406)
(define-constant XK_SCRIPT_SWITCH 65406)
(define-constant XK_NUM_LOCK 65407)
(define-constant XK_KP_SPACE 65408)
(define-constant XK_KP_TAB 65417)
(define-constant XK_KP_ENTER 65421)
(define-constant XK_KP_F1 65425)
(define-constant XK_KP_F2 65426)
(define-constant XK_KP_F3 65427)
(define-constant XK_KP_F4 65428)
(define-constant XK_KP_EQUAL 65469)
(define-constant XK_KP_MULTIPLY 65450)
(define-constant XK_KP_ADD 65451)
(define-constant XK_KP_SEPARATOR 65452)
(define-constant XK_KP_SUBTRACT 65453)
(define-constant XK_KP_DECIMAL 65454)
(define-constant XK_KP_DIVIDE 65455)
(define-constant XK_KP_0 65456)
(define-constant XK_KP_1 65457)
(define-constant XK_KP_2 65458)
(define-constant XK_KP_3 65459)
(define-constant XK_KP_4 65460)
(define-constant XK_KP_5 65461)
(define-constant XK_KP_6 65462)
(define-constant XK_KP_7 65463)
(define-constant XK_KP_8 65464)
(define-constant XK_KP_9 65465)
(define-constant XK_F1 65470)
(define-constant XK_F2 65471)
(define-constant XK_F3 65472)
(define-constant XK_F4 65473)
(define-constant XK_F5 65474)
(define-constant XK_F6 65475)
(define-constant XK_F7 65476)
(define-constant XK_F8 65477)
(define-constant XK_F9 65478)
(define-constant XK_F10 65479)
(define-constant XK_F11 65480)
(define-constant XK_L1 65480)
(define-constant XK_F12 65481)
(define-constant XK_L2 65481)
(define-constant XK_F13 65482)
(define-constant XK_L3 65482)
(define-constant XK_F14 65483)
(define-constant XK_L4 65483)
(define-constant XK_F15 65484)
(define-constant XK_L5 65484)
(define-constant XK_F16 65485)
(define-constant XK_L6 65485)
(define-constant XK_F17 65486)
(define-constant XK_L7 65486)
(define-constant XK_F18 65487)
(define-constant XK_L8 65487)
(define-constant XK_F19 65488)
(define-constant XK_L9 65488)
(define-constant XK_F20 65489)
(define-constant XK_L10 65489)
(define-constant XK_F21 65490)
(define-constant XK_R1 65490)
(define-constant XK_F22 65491)
(define-constant XK_R2 65491)
(define-constant XK_F23 65492)
(define-constant XK_R3 65492)
(define-constant XK_F24 65493)
(define-constant XK_R4 65493)
(define-constant XK_F25 65494)
(define-constant XK_R5 65494)
(define-constant XK_F26 65495)
(define-constant XK_R6 65495)
(define-constant XK_F27 65496)
(define-constant XK_R7 65496)
(define-constant XK_F28 65497)
(define-constant XK_R8 65497)
(define-constant XK_F29 65498)
(define-constant XK_R9 65498)
(define-constant XK_F30 65499)
(define-constant XK_R10 65499)
(define-constant XK_F31 65500)
(define-constant XK_R11 65500)
(define-constant XK_F32 65501)
(define-constant XK_R12 65501)
(define-constant XK_R13 65502)
(define-constant XK_F33 65502)
(define-constant XK_F34 65503)
(define-constant XK_R14 65503)
(define-constant XK_F35 65504)
(define-constant XK_R15 65504)
(define-constant XK_SHIFT_L 65505)
(define-constant XK_SHIFT_R 65506)
(define-constant XK_CONTROL_L 65507)
(define-constant XK_CONTROL_R 65508)
(define-constant XK_CAPS_LOCK 65509)
(define-constant XK_SHIFT_LOCK 65510)
(define-constant XK_META_L 65511)
(define-constant XK_META_R 65512)
(define-constant XK_ALT_L 65513)
(define-constant XK_ALT_R 65514)
(define-constant XK_SUPER_L 65515)
(define-constant XK_SUPER_R 65516)
(define-constant XK_HYPER_L 65517)
(define-constant XK_HYPER_R 65518)
(define-constant XK_SPACE 32)
(define-constant XK_EXCLAM 33)
(define-constant XK_QUOTEDBL 34)
(define-constant XK_NUMBERSIGN 35)
(define-constant XK_DOLLAR 36)
(define-constant XK_PERCENT 37)
(define-constant XK_AMPERSAND 38)
(define-constant XK_QUOTERIGHT 39)
(define-constant XK_PARENLEFT 40)
(define-constant XK_PARENRIGHT 41)
(define-constant XK_ASTERISK 42)
(define-constant XK_PLUS 43)
(define-constant XK_COMMA 44)
(define-constant XK_MINUS 45)
(define-constant XK_PERIOD 46)
(define-constant XK_SLASH 47)
(define-constant XK_0 48)
(define-constant XK_1 49)
(define-constant XK_2 50)
(define-constant XK_3 51)
(define-constant XK_4 52)
(define-constant XK_5 53)
(define-constant XK_6 54)
(define-constant XK_7 55)
(define-constant XK_8 56)
(define-constant XK_9 57)
(define-constant XK_COLON 58)
(define-constant XK_SEMICOLON 59)
(define-constant XK_LESS 60)
(define-constant XK_EQUAL 61)
(define-constant XK_GREATER 62)
(define-constant XK_QUESTION 63)
(define-constant XK_AT 64)
(define-constant XK_A 65)
(define-constant XK_B 66)
(define-constant XK_C 67)
(define-constant XK_D 68)
(define-constant XK_E 69)
(define-constant XK_F 70)
(define-constant XK_G 71)
(define-constant XK_H 72)
(define-constant XK_I 73)
(define-constant XK_J 74)
(define-constant XK_K 75)
(define-constant XK_L 76)
(define-constant XK_M 77)
(define-constant XK_N 78)
(define-constant XK_O 79)
(define-constant XK_P 80)
(define-constant XK_Q 81)
(define-constant XK_R 82)
(define-constant XK_S 83)
(define-constant XK_T 84)
(define-constant XK_U 85)
(define-constant XK_V 86)
(define-constant XK_W 87)
(define-constant XK_X 88)
(define-constant XK_Y 89)
(define-constant XK_Z 90)
(define-constant XK_BRACKETLEFT 91)
(define-constant XK_BACKSLASH 92)
(define-constant XK_BRACKETRIGHT 93)
(define-constant XK_ASCIICIRCUM 94)
(define-constant XK_UNDERSCORE 95)
(define-constant XK_QUOTELEFT 96)
(define-constant XK_LCA 97)
(define-constant XK_LCB 98)
(define-constant XK_LCC 99)
(define-constant XK_LCD 100)
(define-constant XK_LCE 101)
(define-constant XK_LCF 102)
(define-constant XK_LCG 103)
(define-constant XK_LCH 104)
(define-constant XK_LCI 105)
(define-constant XK_LCJ 106)
(define-constant XK_LCK 107)
(define-constant XK_LCL 108)
(define-constant XK_LCM 109)
(define-constant XK_LCN 110)
(define-constant XK_LCO 111)
(define-constant XK_LCP 112)
(define-constant XK_LCQ 113)
(define-constant XK_LCR 114)
(define-constant XK_LCS 115)
(define-constant XK_LCT 116)
(define-constant XK_LCU 117)
(define-constant XK_LCV 118)
(define-constant XK_LCW 119)
(define-constant XK_LCX 120)
(define-constant XK_LCY 121)
(define-constant XK_LCZ 122)
(define-constant XK_BRACELEFT 123)
(define-constant XK_BAR 124)
(define-constant XK_BRACERIGHT 125)
(define-constant XK_ASCIITILDE 126)
(define-constant XK_NOBREAKSPACE 160)
(define-constant XK_EXCLAMDOWN 161)
(define-constant XK_CENT 162)
(define-constant XK_STERLING 163)
(define-constant XK_CURRENCY 164)
(define-constant XK_YEN 165)
(define-constant XK_BROKENBAR 166)
(define-constant XK_SECTION 167)
(define-constant XK_DIAERESIS 168)
(define-constant XK_COPYRIGHT 169)
(define-constant XK_ORDFEMININE 170)
(define-constant XK_GUILLEMOTLEFT 171)
(define-constant XK_NOTSIGN 172)
(define-constant XK_HYPHEN 173)
(define-constant XK_REGISTERED 174)
(define-constant XK_MACRON 175)
(define-constant XK_DEGREE 176)
(define-constant XK_PLUSMINUS 177)
(define-constant XK_TWOSUPERIOR 178)
(define-constant XK_THREESUPERIOR 179)
(define-constant XK_ACUTE 180)
(define-constant XK_MU 181)
(define-constant XK_PARAGRAPH 182)
(define-constant XK_PERIODCENTERED 183)
(define-constant XK_CEDILLA 184)
(define-constant XK_ONESUPERIOR 185)
(define-constant XK_MASCULINE 186)
(define-constant XK_GUILLEMOTRIGHT 187)
(define-constant XK_ONEQUARTER 188)
(define-constant XK_ONEHALF 189)
(define-constant XK_THREEQUARTERS 190)
(define-constant XK_QUESTIONDOWN 191)
(define-constant XK_AGRAVE 192)
(define-constant XK_AACUTE 193)
(define-constant XK_ACIRCUMFLEX 194)
(define-constant XK_ATILDE 195)
(define-constant XK_ADIAERESIS 196)
(define-constant XK_ARING 197)
(define-constant XK_AE 198)
(define-constant XK_CCEDILLA 199)
(define-constant XK_EGRAVE 200)
(define-constant XK_EACUTE 201)
(define-constant XK_ECIRCUMFLEX 202)
(define-constant XK_EDIAERESIS 203)
(define-constant XK_IGRAVE 204)
(define-constant XK_IACUTE 205)
(define-constant XK_ICIRCUMFLEX 206)
(define-constant XK_IDIAERESIS 207)
(define-constant XK_ETH 208)
(define-constant XK_NTILDE 209)
(define-constant XK_OGRAVE 210)
(define-constant XK_OACUTE 211)
(define-constant XK_OCIRCUMFLEX 212)
(define-constant XK_OTILDE 213)
(define-constant XK_ODIAERESIS 214)
(define-constant XK_MULTIPLY 215)
(define-constant XK_OOBLIQUE 216)
(define-constant XK_UGRAVE 217)
(define-constant XK_UACUTE 218)
(define-constant XK_UCIRCUMFLEX 219)
(define-constant XK_UDIAERESIS 220)
(define-constant XK_YACUTE 221)
(define-constant XK_THORN 222)
(define-constant XK_SSHARP 223)
(define-constant XK_LCAGRAVE 224)
(define-constant XK_LCAACUTE 225)
(define-constant XK_LCACIRCUMFLEX 226)
(define-constant XK_LCATILDE 227)
(define-constant XK_LCADIAERESIS 228)
(define-constant XK_LCARING 229)
(define-constant XK_LCAE 230)
(define-constant XK_LCCCEDILLA 231)
(define-constant XK_LCEGRAVE 232)
(define-constant XK_LCEACUTE 233)
(define-constant XK_LCECIRCUMFLEX 234)
(define-constant XK_LCEDIAERESIS 235)
(define-constant XK_LCIGRAVE 236)
(define-constant XK_LCIACUTE 237)
(define-constant XK_LCICIRCUMFLEX 238)
(define-constant XK_LCIDIAERESIS 239)
(define-constant XK_LCETH 240)
(define-constant XK_LCNTILDE 241)
(define-constant XK_LCOGRAVE 242)
(define-constant XK_LCOACUTE 243)
(define-constant XK_LCOCIRCUMFLEX 244)
(define-constant XK_LCOTILDE 245)
(define-constant XK_LCODIAERESIS 246)
(define-constant XK_DIVISION 247)
(define-constant XK_OSLASH 248)
(define-constant XK_LCUGRAVE 249)
(define-constant XK_LCUACUTE 250)
(define-constant XK_LCUCIRCUMFLEX 251)
(define-constant XK_LCUDIAERESIS 252)
(define-constant XK_LCYACUTE 253)
(define-constant XK_LCTHORN 254)
(define-constant XK_YDIAERESIS 255)
(define-constant DXK_RING_ACCENT 268500656)
(define-constant DXK_CIRCUMFLEX_ACCENT 268500574)
(define-constant DXK_CEDILLA_ACCENT 268500524)
(define-constant DXK_ACUTE_ACCENT 268500519)
(define-constant DXK_GRAVE_ACCENT 268500576)
(define-constant DXK_TILDE 268500606)
(define-constant DXK_DIAERESIS 268500514)
(define-constant DXK_REMOVE 268500736)
(define-constant XC_NUM_GLYPHS 154)
(define-constant XC_X_CURSOR 0)
(define-constant XC_ARROW 2)
(define-constant XC_BASED_ARROW_DOWN 4)
(define-constant XC_BASED_ARROW_UP 6)
(define-constant XC_BOAT 8)
(define-constant XC_BOGOSITY 10)
(define-constant XC_BOTTOM_LEFT_CORNER 12)
(define-constant XC_BOTTOM_RIGHT_CORNER 14)
(define-constant XC_BOTTOM_SIDE 16)
(define-constant XC_BOTTOM_TEE 18)
(define-constant XC_BOX_SPIRAL 20)
(define-constant XC_CENTER_PTR 22)
(define-constant XC_CIRCLE 24)
(define-constant XC_CLOCK 26)
(define-constant XC_COFFEE_MUG 28)
(define-constant XC_CROSS 30)
(define-constant XC_CROSS_REVERSE 32)
(define-constant XC_CROSSHAIR 34)
(define-constant XC_DIAMOND_CROSS 36)
(define-constant XC_DOT 38)
(define-constant XC_DOTBOX 40)
(define-constant XC_DOUBLE_ARROW 42)
(define-constant XC_DRAFT_LARGE 44)
(define-constant XC_DRAFT_SMALL 46)
(define-constant XC_DRAPED_BOX 48)
(define-constant XC_EXCHANGE 50)
(define-constant XC_FLEUR 52)
(define-constant XC_GOBBLER 54)
(define-constant XC_GUMBY 56)
(define-constant XC_HAND1 58)
(define-constant XC_HAND2 60)
(define-constant XC_HEART 62)
(define-constant XC_ICON 64)
(define-constant XC_IRON_CROSS 66)
(define-constant XC_LEFT_PTR 68)
(define-constant XC_LEFT_SIDE 70)
(define-constant XC_LEFT_TEE 72)
(define-constant XC_LEFTBUTTON 74)
(define-constant XC_LL_ANGLE 76)
(define-constant XC_LR_ANGLE 78)
(define-constant XC_MAN 80)
(define-constant XC_MIDDLEBUTTON 82)
(define-constant XC_MOUSE 84)
(define-constant XC_PENCIL 86)
(define-constant XC_PIRATE 88)
(define-constant XC_PLUS 90)
(define-constant XC_QUESTION_ARROW 92)
(define-constant XC_RIGHT_PTR 94)
(define-constant XC_RIGHT_SIDE 96)
(define-constant XC_RIGHT_TEE 98)
(define-constant XC_RIGHTBUTTON 100)
(define-constant XC_RTL_LOGO 102)
(define-constant XC_SAILBOAT 104)
(define-constant XC_SB_DOWN_ARROW 106)
(define-constant XC_SB_H_DOUBLE_ARROW 108)
(define-constant XC_SB_LEFT_ARROW 110)
(define-constant XC_SB_RIGHT_ARROW 112)
(define-constant XC_SB_UP_ARROW 114)
(define-constant XC_SB_V_DOUBLE_ARROW 116)
(define-constant XC_SHUTTLE 118)
(define-constant XC_SIZING 120)
(define-constant XC_SPIDER 122)
(define-constant XC_SPRAYCAN 124)
(define-constant XC_STAR 126)
(define-constant XC_TARGET 128)
(define-constant XC_TCROSS 130)
(define-constant XC_TOP_LEFT_ARROW 132)
(define-constant XC_TOP_LEFT_CORNER 134)
(define-constant XC_TOP_RIGHT_CORNER 136)
(define-constant XC_TOP_SIDE 138)
(define-constant XC_TOP_TEE 140)
(define-constant XC_TREK 142)
(define-constant XC_UL_ANGLE 144)
(define-constant XC_UMBRELLA 146)
(define-constant XC_UR_ANGLE 148)
(define-constant XC_WATCH 150)
(define-constant XC_XTERM 152)
